package info.textgrid.middleware;

import java.text.ParseException;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import javax.xml.datatype.DatatypeConfigurationException;
import org.apache.commons.logging.LogFactory;
import org.elasticsearch.action.search.SearchResponse;
import org.elasticsearch.index.query.QueryBuilder;
import org.elasticsearch.index.query.QueryBuilders;
import org.elasticsearch.search.SearchHit;
import org.json.JSONObject;
import info.textgrid.middleware.oaipmh.HeaderType;
import info.textgrid.middleware.oaipmh.ListIdentifiersType;
import info.textgrid.middleware.oaipmh.RequestType;
import info.textgrid.middleware.oaipmh.ResumptionTokenType;

/**
 *
 */
public abstract class IdentifierListDelivererAbstract implements IdentifierListDelivererInterface {

  // **
  // FINALS
  // **

  protected static final int lifeTimeResToken = 100;

  // **
  // STATIC
  // **

  private static org.apache.commons.logging.Log log =
      LogFactory.getLog(RecordListDelivererAbstract.class);

  // **
  // CLASS
  // **

  protected boolean textgrid;
  protected boolean dariah;
  protected boolean foundItems;
  protected long resultSize;
  protected int searchResponseSize;
  protected ResumptionTokenType resTokenForResponse;
  protected String datestamp;
  protected boolean idExist = true;
  protected OAI_ESClient oaiEsClient;
  protected String[] identifierListFields; // Fields for the elasticsearch request
  protected String rangeField; // Field for the optional range query
  protected String dateOfObjectCreation; // Field for the object creation in the repository
  protected String repositoryObjectURIPrefix;
  protected String identifierField;
  protected String formatField;
  protected String formatToFilter;
  protected String specField;
  protected String specFieldPrefix;

  /**
   * @param textgrid
   * @param dariah
   */
  public IdentifierListDelivererAbstract(boolean textgrid, boolean dariah) {
    this.textgrid = textgrid;
    this.dariah = dariah;
  }

  // **
  // PUBLIC METHODS
  // **

  /**
   * @param request
   * @return
   */
  public static ErrorHandler requestChecker(RequestType request) {

    ErrorHandler result = new ErrorHandler();

    // Check for metadata prefix AND resumption token.
    if (request.getMetadataPrefix() != null && request.getResumptionToken() != null) {
      result.setError(OAIPMHConstants.OAI_BAD_ARGUMENT,
          "The resumptionToken is an exclusive argument, please remove metadataPrefix!");
      return result;
    }

    // Check if metadata prefix is existing and valid.
    if (request.getMetadataPrefix() != null
        && !request.getMetadataPrefix().equals(OAIPMHUtilities.OAIDC_PREFIX)
        && !request.getMetadataPrefix().equals(OAIPMHConstants.METADATA_IDIOM_PREFIX)
        && !request.getMetadataPrefix().equals(OAIPMHConstants.METADATA_OPENAIRE_PREFIX)
        && !request.getMetadataPrefix().equals(OAIPMHConstants.METADATA_IDIOM_IMAGE_PREFIX)) {
      result.setError(OAIPMHConstants.OAI_METADATA_FORMAT_ERROR,
          "The value of the metadataPrefix " + request.getMetadataPrefix()
              + " is not supported by the item identified by the value of: "
              + request.getIdentifier());
      return result;
    }

    // Check params in general.
    List<String> errorValues = new ArrayList<String>();

    if (request.getResumptionToken() == null && request.getMetadataPrefix() == null) {
      errorValues.add("metadataPrefix");
    }
    if (request.getIdentifier() != null) {
      errorValues.add("identifier");
    }
    if (errorValues.size() > 0) {
      result.setError(OAIPMHConstants.OAI_BAD_ARGUMENT,
          "The request includes illegal arguments or is missing required arguments: "
              + errorValues);
    }

    return result;
  }

  /**
   * @param listFurtherValues
   * @param lit
   * @param set
   * @param resumptionToken
   * @param cursorCollector
   * @return
   */
  public SearchResponse hitHandling(SearchResponse listFurtherValues, ListIdentifiersType lit,
      String set, String resumptionToken, Map<String, Integer> cursorCollector) {

    int i = 0;
    long size = listFurtherValues.getHits().totalHits;
    setResultSize(size);

    log.debug("hits: " + size);

    for (SearchHit hit : listFurtherValues.getHits().getHits()) {
      i++;

      log.debug("hit " + i + ": " + hit.getId());

      // Handle TextGrid.
      if (this.textgrid) {
        String datestamp = hit.getSourceAsMap().get(this.dateOfObjectCreation).toString();
        try {
          datestamp = OAIPMHUtilities.convertDateFormat(datestamp).toXMLFormat();
        } catch (ParseException e) {
          log.error(e.getMessage());
        } catch (DatatypeConfigurationException e) {
          log.error(e.getMessage());
        }
        String identifier = hit.getSourceAsMap().get(this.identifierField).toString();
        lit = setListIdentifierHeader(datestamp, identifier, lit, set);
      }

      // Handle DARIAH.
      else if (this.dariah) {

        log.debug("dateOfObjectCreation: " + this.dateOfObjectCreation);

        JSONObject json = new JSONObject(hit.getSourceAsMap());

        // Get modifiedDate field and convert datestamp.
        this.datestamp = OAIPMHUtilities.fieldLoader(json, this.dateOfObjectCreation);
        try {
          log.debug("datestamp: " + this.datestamp);

          this.datestamp = OAIPMHUtilities.convertDateFormat(this.datestamp).toXMLFormat();

          log.debug("datestamp converted: " + this.datestamp);

        } catch (ParseException e) {
          log.error(e.getMessage());
          // FIXME Go to ERROR state!
        } catch (DatatypeConfigurationException e) {
          log.error(e.getMessage());
          // FIXME Go to ERROR state!
        }

        // Get identifier field.
        String identifier = OAIPMHUtilities.fieldLoader(json, this.identifierField);

        log.debug("identifier: " + identifier);

        lit = setListIdentifierHeader(this.datestamp, identifier, lit, set);
      }
    }

    // Check the need for a resumption token!
    ResumptionTokenType resTokenForResponse = OAIPMHUtilities.getResumptionToken(
        listFurtherValues.getHits().getTotalHits(), resumptionToken, cursorCollector,
        listFurtherValues.getScrollId(), this.searchResponseSize, i);
    if (resTokenForResponse != null) {
      lit.setResumptionToken(resTokenForResponse);
    }

    return listFurtherValues;
  }

  /**
   * @param datestamp
   * @param identifier
   * @param lit
   * @param set
   * @return
   */
  private ListIdentifiersType setListIdentifierHeader(String datestamp, String identifier,
      ListIdentifiersType lit, String set) {

    HeaderType header = new HeaderType();
    header.setDatestamp(datestamp);
    header.setIdentifier(identifier);

    if (set != null) {
      header.getSetSpec().add(this.specFieldPrefix + set);
    }
    lit.getHeader().add(header);

    return lit;
  }

  /**
   * @param set
   * @param from
   * @param until
   * @return
   */
  public QueryBuilder setOrNot(String set, String from, String until) {

    QueryBuilder rangeQuery = QueryBuilders.rangeQuery(this.rangeField).from(from).to(until);
    QueryBuilder formatQuery = QueryBuilders.matchPhrasePrefixQuery("format", this.formatToFilter);
    QueryBuilder query;

    if (set != null) {
      String queryField = "";
      String valueField = "";

      if (this.dariah == true) {
        queryField = "descriptiveMetadata.dc:relation";
        valueField = set;
      }

      if (this.textgrid == true) {
        String[] setParts = set.split(":");
        if (setParts[0].equals("project")) {
          queryField = "project.id";
          valueField = setParts[1];
        }
      }

      QueryBuilder matchQuery = QueryBuilders.matchPhraseQuery(queryField, valueField);
      if (this.dariah) {
        query = matchQuery;
      } else {
        query = QueryBuilders.boolQuery().must(matchQuery).must(formatQuery).must(rangeQuery);
      }

    } else {
      if (this.dariah == true)
        query = rangeQuery;
      else
        query = QueryBuilders.boolQuery().must(rangeQuery).must(formatQuery);
    }

    return query;
  }

  // **
  // GETTERS & SETTERS
  // **

  /**
   * @return
   */
  public boolean isFoundItems() {
    return this.foundItems;
  }

  /**
   * @param foundItems
   */
  public void setFoundItems(boolean foundItems) {
    this.foundItems = foundItems;
  }

  /**
   * @return
   */
  public long getResultSize() {
    return this.resultSize;
  }

  /**
   * @param resultSize
   */
  public void setResultSize(long resultSize) {
    this.resultSize = resultSize;
  }

  /**
   * @return
   */
  public ResumptionTokenType getResTokenForResponse() {
    return this.resTokenForResponse;
  }

  /**
   * @param resTokenForResponse
   */
  public void setResTokenForResponse(ResumptionTokenType resTokenForResponse) {
    this.resTokenForResponse = resTokenForResponse;
  }

  /**
   * @return
   */
  public String getFieldForRange() {
    return this.rangeField;
  }

  /**
   * @param fieldForRange
   */
  public void setFieldForRange(String fieldForRange) {
    this.rangeField = fieldForRange;
  }

  /**
   * @return
   */
  public OAI_ESClient getOaiEsClient() {
    return this.oaiEsClient;
  }

  /**
   * @param oaiEsClient
   */
  public void setOaiEsClient(OAI_ESClient oaiEsClient) {
    this.oaiEsClient = oaiEsClient;
  }

  /**
   * @param lit
   * @return
   */
  public ListIdentifiersType getLit(ListIdentifiersType lit) {
    return lit;
  }

  /**
   * @return
   */
  public String getDateOfObjectCreation() {
    return this.dateOfObjectCreation;
  }

  /**
   * @param dateOfObjectCreation
   */
  public void setDateOfObjectCreation(String dateOfObjectCreation) {
    this.dateOfObjectCreation = dateOfObjectCreation;
  }

  /**
   * @param repositoryObjectURIPrefix
   */
  public void setRepositoryObjectURIPrefix(String repositoryObjectURIPrefix) {
    this.repositoryObjectURIPrefix = repositoryObjectURIPrefix;
  }

  /**
   * @return
   */
  public String getRepositoryObjectURIPrefix() {
    return this.repositoryObjectURIPrefix;
  }

  /**
   * @param validItem
   */
  public void setIdExist(boolean validItem) {
    this.idExist = validItem;
  }

  /**
   * @return
   */
  public boolean getIdExist() {
    return this.idExist;
  }

  /**
   * @return
   */
  public String[] getIdentifierListFields() {
    return this.identifierListFields;
  }

  /**
   * @param identifierListFields
   */
  public void setIdentifierListFields(String[] identifierListFields) {
    this.identifierListFields = identifierListFields;
  }

  /**
   * @return
   */
  public String getFormatField() {
    return this.formatField;
  }

  /**
   * @param formatField
   */
  public void setFormatField(String formatField) {
    this.formatField = formatField;
  }

  /**
   * @return
   */
  public String getFormatToFilter() {
    return this.formatToFilter;
  }

  /**
   * @param formatToFilter
   */
  public void setFormatToFilter(String formatToFilter) {
    this.formatToFilter = formatToFilter;
  }

  /**
   * @return
   */
  public String getIdentifierField() {
    return this.identifierField;
  }

  /**
   * @param identifierPrefix
   */
  public void setIdentifierField(String identifierPrefix) {
    this.identifierField = identifierPrefix;
  }

  /**
   * @return
   */
  public int getSearchResponseSize() {
    return this.searchResponseSize;
  }

  /**
   * @param searchResponseSize
   */
  public void setSearchResponseSize(String searchResponseSize) {
    this.searchResponseSize = Integer.parseInt(searchResponseSize);
  }

  /**
   * @return
   */
  public String getSpecField() {
    return this.specField;
  }

  /**
   * @param specField
   */
  public void setSpecField(String specField) {
    this.specField = specField;
  }

  /**
   * @return
   */
  public String getSpecFieldPrefix() {
    return this.specFieldPrefix;
  }

  /**
   * @param specFieldPrefix
   */
  public void setSpecFieldPrefix(String specFieldPrefix) {
    this.specFieldPrefix = specFieldPrefix;
  }

}
